#!/usr/bin/python

#    Copyright 2004 Andrew Wilkinson <aw@cs.york.ac.uk>.
#
#    This file is part of PyLinda (http://www-users.cs.york.ac.uk/~aw/pylinda)
#
#    PyLinda is free software; you can redistribute it and/or modify
#    it under the terms of the GNU Lesser General Public License as published by
#    the Free Software Foundation; either version 2.1 of the License, or
#    (at your option) any later version.
#
#    PyLinda is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU Lesser General Public License for more details.
#
#    You should have received a copy of the GNU Lesser General Public License
#    along with PyLinda; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

import socket
import SocketServer
import struct
import sys
import thread

#from linda.profile import P

from linda.messages import *

from linda.options import getOptions
from linda.tscontainer import TupleSpaceContainer
from linda.tuplespace import TupleSpace
from linda.stats import Stats
stats = Stats()

import linda.kernel as kernel
kernel.run_as_server = True

import linda.utils as utils

node_id = 1
neighbours = {}

getMsgId = utils.Counter()
process_id = utils.Counter()
ts_ids = utils.Counter()

processes = {}

local_ts = TupleSpaceContainer()
blocked_processes = {}

connections = {}

class LindaConnection(SocketServer.BaseRequestHandler):
    def handle(self):
        print "new connection from " + str(self.client_address)
        self.request.setblocking(1)

        self.messages = {
            register_process: self.register_process,
            unregister_process: self.unregister_process,
            my_name_is: self.my_name_is,
            create_tuplespace: self.create_tuplespace,
            get_connect_details: self.get_connect_details,
            get_new_node_id: self.get_new_node_id,
            get_node_id: self.get_node_id,
            read_tuple: self.read_tuple,
            in_tuple: self.in_tuple,
            out_tuple: self.out_tuple,
            unblock: self.unblock,
            return_tuple: self.return_tuple,
            collect: self.collect,
            copy_collect: self.copy_collect,
            multiple_in: self.multiple_in,
            increment_ref: self.increment_ref,
            decrement_ref: self.decrement_ref,
            get_references: self.get_references,
            get_neighbours: self.get_neighbours,
            get_blocked_list: self.get_blocked_list,
            get_stats: self.get_stats,
            kill_server: self.kill_server,
            }

        # check that the connection comes from an allowed source
        if self.server.allowed_peers:
            ok = False
            for addr in self.server.allowed_peers:
                if self.verify_address(addr):
                    ok = True
                    break
            if not ok:
                print "DENIED ACCESS FROM", self.client_address
                self.request.shutdown(2) # this shuts the connection down but the other end doesn't notice
                self.request.close()
                return

        self.other_pid = None
        self.other_nid = None
        self.semaphore = threading.Semaphore()

        # Here we decide if we can use a better form of communication
        while True:
            message = utils.recv(self.request)
            m = utils.decode(message)
            if m[2] == begin_session:
                utils.send(self.request, done)
                break

            elif m[2] == support_domain:
                try:
                    if options.use_domain:
                        import domain_socket
                    else:
                        raise ImportError
                except ImportError:
                    utils.send(self.request, dont_know)
                else:
                    s = domain_socket.Socket()
                    s.create()
                    utils.send(self.request, utils.encode(s.getaddr()))
                    s.accept()
                    self.request.shutdown(1)
                    self.request.close()
                    self.request = s
            else:
                print "Unknown Session Setup Message: %s" % (m[2], )
                utils.send(self.request, dont_know)

        connections[thread.get_ident()] = self.request

        while True:
            self.semaphore.release()
            try:
                message = utils.recv(self.request)
            except socket.error, msg:
                # should we do something other than silently pass this error?
                print "Got A Socket Error - " + str(msg)
                message = ""
            self.semaphore.acquire()

            # when a connection is broken a zero length string is returned from self.request.recv
            if not message:
                print "connection to %s (%s) being broken" % (str(self.client_address), str(self.other_pid or self.other_nid))
                break

            try:
                m = utils.decode(message)
            except:
                print repr(message)
                raise

            if m[2] == close_connection:
                 utils.send(self.request, ok)
                 self.request.close()
                 break

            #P.enter(m[2])
            self.handle_msg(m[0], m[1], m[2], m[3:])
            #P._exit()

            del m

        del connections[thread.get_ident()]

        self.request.close()

        # Once a connection is broken we end up here
        if self.other_pid is not None:
            # When a process disconnects ensure that all their references are removed
            removeProcess(node_id, getMsgId(), self.other_pid)
            if self.other_pid != kernel.process_id: # don't count the loopback
                stats.userDisconnect() # update stats
        elif self.other_nid is not None:
            stats.serverDisconnect() # update stats

    def verify_address(self, addr):
        flds = addr.split("/")
        addr = flds[0]
        if len(flds) == 1:
            mask = 32
        elif len(flds) == 2:
            mask = int(flds[1])

        addr = struct.unpack("!i", socket.inet_aton(addr))[0] & utils.mask(mask)
        client = struct.unpack("!i", socket.inet_aton(self.client_address[0]))[0] & utils.mask(mask)

        return addr == client

    def handle_msg(self, node, mid, message, data):
        try:
            self.messages[message](node, mid, message, data)
        except KeyError:
            print "Unknown Message: %s (%s)" % (message, str(data))
            utils.send(self.request, done)
    
    def get_new_node_id(self, node, mid, message, data):
        # when a server connects to the network they want a new node id
        if node == 0:
            node_ids = broadcast_message(node_id, getMsgId(), get_node_id, data[0])
        else:
            node_ids = broadcast_message(node, mid, get_node_id, data[0])
  
        new_id = max(map(int, node_ids + [node_id]))+1
        print str(get_new_node_id)+" : "+str(new_id)
        utils.send(self.request, str(new_id))
  
        # the connecting computer has now got an id, add them to our neighbours
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((self.client_address[0], data[0]))
  
        neighbours[new_id] = ((new_id, s, data[0], threading.Semaphore()))
  
        utils.send(s, utils.encode((0,0,begin_session)))
        r = utils.recv(s)
                
        self.other_nid = new_id

    def get_node_id(self, node, mid, message, data):
                utils.send(self.request, str(node_id))

    def my_name_is(self, node, mid, message, data):
                # when someone connects who already has an id they need to let us know who they are...
                if utils.isNodeId(data[0]):
                    self.other_nid = data[0]
                    stats.serverConnect() # update stats
                else:
                    self.other_pid = data[0]
                    stats.userConnect() # update stats

                utils.send(self.request, done)

    def register_process(self, node, mid, message, data):
                # When a new process connects they need to acquire new process id
                p_id = "%i!%i" % (node_id, process_id.next())
                processes[p_id] = self.request
                self.other_pid = p_id
                stats.userConnect() # update stats

                utils.send(self.request, p_id)

    def unregister_process(self, node, mid, message, data):
                # if a process is about to disconnect they can let us know first
                # removeProcess removes any references held by the process
                removeProcess(node, mid, data[0], data[1])

                utils.send(self.request, done)

    def get_connect_details(self, node, mid, message, data):
                try:
                   int(data[0])
                except ValueError:
                   print "Error: ", data[0]
                   raise
                # try to find out how to connect to a server
                if int(data[0]) == node_id:
                    # they're looking for us! Return our details
                    utils.send(self.request, utils.encode(self.request.getsockname()))
                elif int(data[0]) in neighbours.keys():
                    # they're looking for our neighbour, ask them how to connect to them
                    utils.send(self.request, sendMessageToNode(int(data[0]), node, mid, get_connect_details, int(data[0])))
                else:
                    # we don't know the node they're looking for
                    utils.send(self.request, dont_know)

    def create_tuplespace(self, node, mid, message, data):
                # return a new tuplespace id
                ts = "%i:%i" % (node_id, ts_ids.next())

                local_ts.newTupleSpace(ts)
                utils.send(self.request, ts)

    def out_tuple(self, node, mid, message, data):
                # output a tuple into a tuplespace
                ts, tup = data
                if local_ts.has_key(ts):
                    # we own the given tuplespace - drop the tuple into it
                    tup = utils.decode(tup)
                    utils.containsTS(tup, lambda t: utils.changeOwner(t, ts))

                    local_ts[ts]._out(tup)
                    stats.out_tuple() # update stats

                    utils.send(self.request, done)
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, out_tuple, ts, tup))

    def read_tuple(self, node, mid, message, data):
                ts, template, pid, unblockable = data

                if self.other_pid is not None:
                    blocked_processes[self.other_pid] = (self.request, self.semaphore)

                if local_ts.has_key(ts):
                    r = local_ts[ts]._rd(pid, template, unblockable)

                    if r is not None and self.other_pid is not None:
                        del blocked_processes[self.other_pid]
                        utils.send(self.request, r)
                    elif r is None and self.other_pid is None:
                        # if we're talking to another server tell them we're done
                        utils.send(self.request, done)
                    elif r is not None and self.other_pid is None:
                        # if we're talking to another server tell them we're done
                        utils.send(self.request, r)
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    r = sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, read_tuple, ts, template, pid, unblockable)
                    if r != done:
                        utils.send(self.request, r)

    def in_tuple(self, node, mid, message, data):
                ts, template, pid, unblockable = data

                if self.other_pid is not None:
                    blocked_processes[self.other_pid] = (self.request, self.semaphore)

                if local_ts.has_key(ts):
                    r = local_ts[ts]._in(pid, template, unblockable)
                    stats.in_tuple() # update stats

                    if r is not None and self.other_pid is not None:
                        del blocked_processes[self.other_pid]
                        utils.send(self.request, r)
                    elif r is None and self.other_pid is None:
                        # if we're talking to another server tell them we're done
                        utils.send(self.request, done)
                    elif r is not None and self.other_pid is None:
                        # if we're talking to another server tell them we're done
                        utils.send(self.request, r)
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    r = sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, in_tuple, ts, template, pid, unblockable)
                    if r != done:
                        utils.send(self.request, r)

    def return_tuple(self, node, mid, message, data):
                pid, tup = data

                if pid in blocked_processes.keys():
                    s, semaphore = blocked_processes[pid]
                    del blocked_processes[pid]
                    semaphore.acquire()
                    utils.send(s, tup)
                    semaphore.release()
                    utils.send(self.request, done)
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    utils.send(self.request, sendMessageToNode(utils.getNodeFromProcessId(pid), node, mid, return_tuple, pid, tup))

    def unblock(self, node, mid, message, data):
                pid = data[0]

                if pid in blocked_processes.keys():
                    s, semaphore = blocked_processes[pid]
                    del blocked_processes[pid]
                    semaphore.acquire()
                    utils.send(s, unblock)
                    semaphore.release()
                    utils.send(self.request, done)
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    utils.send(self.request, sendMessageToNode(utils.getNodeFromProcessId(pid), node, mid, unblock, pid))

    def collect(self, node, mid, message, data):
                ts, dest_ts, template = data
                if local_ts.has_key(ts):
                    tups = local_ts[ts].collect(template)
                    if tups != []:
                        if local_ts.has_key(dest_ts):
                            for t in tups:
                               utils.containsTS(t, lambda x: utils.changeOwner(x, dest_ts))
                               local_ts[dest_ts]._out(t)
                        else:
                            for t in tups:
                               utils.containsTS(t, lambda x: x._addreference(node_id))
                            sendMessageToNode(utils.getNodeFromTupleSpaceId(dest_ts), node_id, getMsgId(), multiple_in, dest_ts, utils.encode(tups))
                    utils.send(self.request, str(len(tups)))
                else:
                    utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, copy_collect, ts, dest_ts, template))

    def copy_collect(self, node, mid, message, data):
                 ts, dest_ts, template = data
                 if local_ts.has_key(ts):
                     tups = local_ts[ts].copy_collect(template)
                     if tups != []:
                         if local_ts.has_key(dest_ts):
                             for t in tups:
                                 utils.containsTS(t, utils.changeOwner)
                                 local_ts[dest_ts]._out(t)

                         else:
                            for t in tups:
                               utils.containsTS(t, lambda x: x._addreference(node_id))
                            sendMessageToNode(utils.getNodeFromTupleSpaceId(dest_ts), node_id, getMsgId(), multiple_in, dest_ts, utils.encode(tups))

                     utils.send(self.request, str(len(tups)))
                 else:
                     utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, copy_collect, ts, dest_ts, template))

    def multiple_in(self, node, mid, message, data):
                 ts, tups = data
                 if local_ts.has_key(ts):
                     tups = utils.decode(tups)
                     for t in tups:
                         local_ts[ts]._out(t)

                         utils.containsTS(tup, lambda x: utils.changeOwner(x, ts))

                     utils.send(self.request, done)
                 else:
                     utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, multiple_in, ts, tups))

    def increment_ref(self, node, mid, message, data):
                ts, ref = data
                if local_ts.has_key(ts):
                    local_ts.addReference(ts, ref)

                    utils.send(self.request, done)
                else:
                    utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, increment_ref, ts, ref))

    def decrement_ref(self, node, mid, message, data):
                ts, ref = data
                if ts != "0:0":
                    if local_ts.has_key(ts):
                        ts_obj = local_ts[ts]

                        # see the note in the TupleSpace.removeallreference for an explaination of the killlock
                        ts_obj.killlock.acquire()

                        threading.Thread(target=local_ts.deleteReference,args=(ts, ref)).start()

                        utils.send(self.request, done)
                    else:
                        utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, decrement_ref, ts, ref))

    def get_references(self, node, mid, message, data):
                ts = data[0]
                if utils.getNodeFromTupleSpaceId(ts) == node_id:
                    if local_ts.has_key(ts):
                        ts_obj = local_ts[ts]
                        ts_obj.ref_semaphore.acquire()
                        try:
                            utils.send(self.request, utils.encode(ts_obj.refs))
                        finally:
                            ts_obj.ref_semaphore.release()
                    else:
                        utils.send(self.request, utils.encode([]))
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    utils.send(self.request, sendMessageToNode(utils.getNodeFromTupleSpaceId(ts), node, mid, get_references, ts))

    def get_blocked_list(self, node, mid, message, data):
                ts = data[0]
                if utils.getNodeFromTupleSpaceId(ts) == node_id:
                    if local_ts.has_key(ts):
                        ts_obj = local_ts[ts]
                        ts_obj.blocked_semaphore.acquire()
                        try:
                            utils.send(self.request, utils.encode(ts_obj.blocked_list.keys()))
                        finally:
                            ts_obj.blocked_semaphore.release()
                    else:
                        utils.send(self.request, utils.encode([]))
                else:
                    if node == 0:
                        node, mid = node_id, getMsgId()

                    utils.send(self.request, sendMessageToNode(int(ts.split(":")[0]), node, mid, get_blocked_list, ts))

    def get_neighbours(self, node, mid, message, data):
                utils.send(self.request, utils.encode(neighbours.keys()))

    def get_stats(self, node, mid, message, data):
                utils.send(self.request, stats.getStats())

    def kill_server(self, node, mid, message, data):
                server.close = True
                utils.send(self.request, done)

def broadcast_message(*args):
    memo = [node_id]
    r = []
    todo = neighbours.keys()
    
    while len(todo) > 0:
        node, todo = todo[0], todo[1:]
        if node in memo:
            continue
        else:
            memo.append(node)

        m = sendMessageToNode(node, *args)

        if m != dont_know:
            r.append(m)

        n = sendMessageToNode(node, node_id, getMsgId(), get_neighbours)
        if n != dont_know:
            todo.extend(utils.decode(n))
    return r

def broadcast_firstreplyonly(*args):
    memo = [node_id]
    todo = neighbours.keys()
    while len(todo) > 0:
        node, todo = todo[0], todo[1:]
        if node in memo:
            continue
        else:
            memo.append(node)

        m = sendMessageToNode(node, *args)

        if m != dont_know:
            return m

        n = sendMessageToNode(node, node_id, getMsgId(), get_neighbours)
        if n != dont_know:
            todo.extend(utils.decode(n))

    return dont_know

class LindaServer(SocketServer.ThreadingTCPServer):
    """\internal
    A simple class the implements a threaded socket server - using the ThreadingTCPServer class provided by Python
    """
    def __init__(self, address, handler, allowed_peers=[]):
        """\internal
        """
        SocketServer.ThreadingTCPServer.__init__(self, address, handler)
        self.daemon_threads = True
        self.allowed_peers = allowed_peers
        self.close = False

    def serve_forever(self):
        """Handle one request at a time until doomsday."""
        while not self.close:
            self.handle_request()

    def handle_request(self):
        """Handle one request, possibly blocking."""
        try:
            self.socket.settimeout(10.0)
            request, client_address = self.get_request()
            self.socket.settimeout(None)
        except socket.error:
            return
        if self.verify_request(request, client_address):
            try:
                self.process_request(request, client_address)
            except:
                self.handle_error(request, client_address)
                self.close_request(request)

def sendMessageToNode(node, *args):
    """\internal
    Helper function to ensure that we have connected to the server we want to talk to, and to send the message.
    """
    if node == 0:
        node = 1

    if node == node_id:
        return kernel.message(args)

    if not neighbours.has_key(node):
        connectTo(node)

    neighbours[node][3].acquire()
    try:
        utils.send(neighbours[node][1], utils.encode(args))
        return utils.recv(neighbours[node][1])
    finally:
        neighbours[node][3].release()

def connectTo(node):
    """\internal
    \brief Find the address of, and connect to a new server.
    """
    details = broadcast_firstreplyonly(node_id, getMsgId(), get_connect_details, node)
    details = utils.decode(details)
    
    s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    s.connect(details)
        
    utils.send(s, utils.encode((0,0,begin_session)))
    r = utils.recv(s)
    
    neighbours[node] = (node, s, details[1], threading.Semaphore())

    sendMessageToNode(node, node_id, getMsgId(), my_name_is, node_id)

def removeProcess(node, mid, pid, local=True):
    """\internal
    \brief Called if a process leaves the system to ensure that no references that the process had remain in the system.

    If a process exits normally then all this should be unnessicary, however we can't trust user processes.
    """
    # check it actually is a process and not another node
    if not utils.isProcessId(pid):
        return

    # check that it wasn't blocked when the connection was lost
    if pid in blocked_processes.keys():
       del blocked_processes[pid]

    # remove any references the process may have had to our processes
    for ts in local_ts:
        local_ts.deleteAllReferences(ts, pid)

    # if the process was connected to us then broadcast the fact that it has left the system
    if local:
        broadcast_message(node, mid, unregister_process, pid, False)

import threading
class KernelImport(threading.Thread):
    """\internal
    We need to create a loopback connection, but we can't do that until the server has started - however we can't 
    do it after the server has been started - so we do it in another thread!
    """
    def run(self):
        kernel.connect(options.port)
        kernel.process_id = node_id
        kernel.use_domain = options.use_domain
        kernel.message(my_name_is, node_id)

def main(connection_class = LindaConnection):
    """\internal
    \brief Parse command line options and start the server.
    """
    global server, node_id, neighbours, local_ts, options

    options = getOptions()

    if options.peer:
        options.peer.append("127.0.0.1") # always allow local connections.

    def lookupname(addr):
        try:
            addr,r = addr.split("/")
        except ValueError:
            r = "0"

        addr = socket.gethostbyname(addr)
        if addr.count(".") != 3:
            print "%s is not in n.n.n.n[/r] format" % (addr+"/"+r)
            sys.exit(0)
        return addr+"/"+r

    server = LindaServer((options.bindaddress, options.port),
                          connection_class,
                          map(lookupname, options.peer))

    if options.connect != "":
        s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        s.connect((options.connect, options.connectport))

        utils.send(s, utils.encode((0,0,begin_session)))
        r = utils.recv(s)
        
        if r == "":
            print "Connection Failed: Probably denied by the other server"
            return
        
        utils.send(s, utils.encode((0,0,get_node_id)))
        node = int(utils.recv(s))
        
        utils.send(s, utils.encode((0,0,get_new_node_id,options.port)))
        r = utils.recv(s)
        node_id = int(r)
        print "got my node id: ", node_id

        neighbours[node] = ((node, s, options.connectport, threading.Semaphore()))
    else:
        local_ts.newTupleSpace("0:0")
      
    # import the kernel
    KernelImport().start()

    server.serve_forever()

if __name__ == "__main__":
    try:
        main()
    except KeyboardInterrupt:
        print "KeyboardInterrupt"

    # shut down the kernel's connection to the server
    kernel.close()

    for c in connections.values():
        c.close()
